from ansibler.args.cmd import ARGS


def display_help() -> None:
    """
    Displays some info about the parameters
    """
    args_str = "\n".join([
        f"\t--{arg.get('arg_name')}: {arg.get('arg_help')}"
        for arg in ARGS
    ])
    print(f"You must pass any of the following arguments:\n{args_str}")



from unittest import TestCase
from unittest.mock import patch
from ansibler.role_dependencies.galaxy import get_from_ansible_galaxy
from ansibler.exceptions.ansibler import RoleMetadataError


class TestGalaxy(TestCase):
    @patch("ansibler.role_dependencies.galaxy.get_subprocess_output")
    def test_get_from_galaxy(self, mock_get_subprocess_output):
        mock_out = "description: Ensures Snap is installed"
        mock_get_subprocess_output.return_value = mock_out

        role = get_from_ansible_galaxy("professormanhattan.snapd")
        expected = {
            "namespace": "professormanhattan",
            "role_name": "snapd",
            "description": "Ensures Snap is installed",
            "repository": None,
            "repository_status": None
        }

        self.assertDictEqual(role, expected)

    @patch("ansibler.role_dependencies.galaxy.get_subprocess_output")
    def test_get_from_galaxy_role_not_found(self, mock_get_subprocess_output):
        mock_get_subprocess_output.return_value = ""
        with self.assertRaises(RoleMetadataError):
            _ = get_from_ansible_galaxy("professormanhattan.snapd")



from unittest import TestCase
import pathlib
import shutil
from ansibler.utils.files import (
    check_folder_exists, copy_file, create_folder_if_not_exists, list_files
)


class TestFileUtils(TestCase):
    def setUp(self) -> None:
        """
        Test case setup
        """
        self.example_path = "./test/test_utils/example/"
        self.should_exist_path = "./test/test_utils/exists/"
        self.files_path = "./test/test_utils/example_files/"
        self.copy_path = "./test/test_utils/copy_files/"

    def test_create_folder_if_not_exists(self):
        """
        Test folder creation
        """
        create_folder_if_not_exists(self.example_path)
        self.assertTrue(pathlib.Path(self.example_path).is_dir())
        shutil.rmtree(self.example_path)

    def test_folder_exists(self):
        """
        Test check folder exists
        """
        create_folder_if_not_exists(self.should_exist_path)
        self.assertTrue(check_folder_exists(self.should_exist_path))
        shutil.rmtree(self.should_exist_path)

    def test_folder_not_exists(self):
        """
        Asserts a folder does not exist
        """
        self.assertFalse(check_folder_exists(self.should_exist_path))

    def test_list_files(self):
        """
        Test list files
        """
        # Create dir
        pathlib.Path(self.files_path).mkdir(parents=True, exist_ok=True)
        pathlib.Path(self.files_path + "example.file").touch(exist_ok=True)

        files = list_files(self.files_path)
        self.assertEqual(len(files), 1)

        shutil.rmtree(self.files_path)

    def test_copy_file(self):
        """
        Test copy file
        """
        src = self.copy_path + "example.file"
        dst = self.copy_path + "example.copy"
        # Create dir
        pathlib.Path(self.copy_path).mkdir(parents=True, exist_ok=True)
        pathlib.Path(src).touch(exist_ok=True)

        # Copy
        copy_file(src, dst)

        # Read contents
        src_content, dst_content = None, None
        with open(src) as f:
            src_content = f.read()

        with open(dst) as f:
            dst_content = f.read()

        self.assertEqual(src_content, dst_content)

        shutil.rmtree(self.copy_path)

    def test_copy_file_new_content(self):
        """
        Tests copy file but then changes the dst file contents.
        """
        src = self.copy_path + "example.file"
        dst = self.copy_path + "example.copy"
        # Create dir
        pathlib.Path(self.copy_path).mkdir(parents=True, exist_ok=True)
        pathlib.Path(src).touch(exist_ok=True)

        # Copy
        copy_file(src, dst, new_content="Hello, world!")

        # Read contents
        src_content, dst_content = None, None
        with open(src) as f:
            src_content = f.read()

        with open(dst) as f:
            dst_content = f.read()

        self.assertNotEqual(src_content, dst_content)
        self.assertEqual(dst_content, "Hello, world!")

        shutil.rmtree(self.copy_path)



from unittest import TestCase
from ansibler.role_dependencies.role_info import (
    get_role_full_path,
    get_role_name,
    get_role_name_from_req_file,
)


class TestRoleInfo(TestCase):
    def test_get_role_name(self):
        """
        Test get role name
        """
        role = get_role_name(
            "/home/user/projects/Playbooks/roles/system",
            "/home/user/projects/Playbooks/roles/system/snapd/meta/main.yml"
        )
        self.assertEqual(role, "snapd")

    def test_get_role_name_from_requirements_file(self):
        """
        Test get role name from requirements.yml
        """
        role = get_role_name_from_req_file(
            "/home/user/projects/Playbooks/roles/system",
            "/home/user/projects/Playbooks/roles/system/snapd/requirements.yml"
        )
        self.assertEqual(role, "snapd")

    def test_get_role_full_path(self):
        """
        Test get role full path
        """
        role_path = get_role_full_path(
            "/home/user/projects/Playbooks/roles/system", "snapd")
        self.assertEqual(
            role_path, "/home/user/projects/Playbooks/roles/system/snapd")



from unittest import TestCase
from unittest.mock import patch
from ansibler.molecule_test.parse import (
    parse_play_name,
    parse_play_recap,
    parse_play_recap_dump,
    parse_os,
    parse_recap_value,
    OK_COUNT_PATTERN
)


class TestParseMolecule(TestCase):
    def test_parse_play_name(self):
        """
        Test extract play name
        """
        dump = "\n\nINFO     Running docker-snap >  idempotence\n\n"
        res = parse_play_name(dump)
        self.assertEqual(res, "idempotence")

    @patch("ansibler.molecule_test.parse.parse_play_recap_dump")
    def test_parse_play_recap(self, mock_parse_play_recap_dump):
        """
        Test parse play recap

        Args:
            mock_parse_play_recap_dump (Mock): parse play recap dump mock 
        """
        play_recap = "Debian-10: ok=16 changed=0 unreachable=0 failed=0 " \
                     "skipped=4 rescued=0 ignored=0\nUbuntu-20.04: ok=16 " \
                     "changed=0 unreachable=0 failed=0 skipped=4 rescued=0 " \
                     "ignored=0"
        mock_parse_play_recap_dump.return_value = play_recap

        res = parse_play_recap(play_recap)
        expected_recap = {
            "ok": 16,
            "changed": 0,
            "unreachable": 0,
            "failed": 0,
            "skipped": 4,
            "rescued": 0,
            "ignored": 0
        }
        expected_recap = [
            {"os_name": "Debian", "os_version": "10", **expected_recap},
            {"os_name": "Ubuntu", "os_version": "20.04", **expected_recap}
        ]

        self.assertEqual(res, expected_recap)

    def test_parse_play_recap_dump(self):
        """
        Test parse play recap dump
        """
        dump = """
        changed: [Ubuntu-21.04]
        PLAY RECAP ******
        expected result

        INFO     Running docker-snap > idempotence
        """
        res = parse_play_recap_dump(dump)
        self.assertEqual(res, "expected result")

    def test_parse_os(self):
        """
        Test parse os
        """
        name, version = parse_os("Debian-10    : ok=19")
        self.assertEqual(f"{name}-{version}", "Debian-10")

    def test_parse_os_no_version(self):
        """
        Test parse os no version
        """
        name, version = parse_os("macos    : ok=19")
        self.assertEqual(f"{name}-{version}", "macos-None")

    def test_parse_recap_value(self):
        """
        Test parse recap value
        """
        res = parse_recap_value(OK_COUNT_PATTERN, "Debian-10: ok=19")
        self.assertEqual(res, 19)

    def test_parse_recap_value_not_exists(self):
        """
        Assert returns -1 when the pattern doesnt return a match
        """
        res = parse_recap_value(OK_COUNT_PATTERN, "")
        self.assertEqual(res, -1)



from unittest import TestCase
from unittest.mock import patch
import argparse
from ansibler.args import cmd
from ansibler.args.cmd import configure_parser, validate_arg, read_parser_args


class TestArguments(TestCase):
    VALID_ARGUMENTS = [
        {
            "arg_name": "generate-compatibility-chart",
            "arg_help": "help",
            "arg_action": "store_true"
        },
        {
            "arg_name": "populate-platforms",
            "arg_help": "help",
            "arg_action": "store_true",
        },
        {
            "arg_name": "role-dependencies",
            "arg_help": "help",
            "arg_action": "store_true"
        },
        { "arg_name": "version", "arg_help": "help" },
    ]

    INVALID_ARGUMENTS = [
        { "name": "invalid", "help": "invalid help" }
    ]

    def setUp(self) -> None:
        """
        Test case setup
        """
        self.mock_valid_args = patch.object(
            cmd, 'ARGS', return_value=self.VALID_ARGUMENTS
        )
        with self.mock_valid_args:
            self.parser = configure_parser()

    def test_parser_properly_setup(self):
        """
        Tests initialization of parser
        """
        self.assertIsInstance(self.parser, argparse.ArgumentParser)

    def test_parser_improper_setup(self):
        """
        Tests parser fails to initialize when ARGS is not properly set
        ([{"arg_name": "foo", "arg_help": "bar"}, ...}])
        """
        with patch.object(cmd, "ARGS", self.INVALID_ARGUMENTS):
            with self.assertRaises(ValueError):
                _ = configure_parser()

    def test_valid_argument(self):
        """
        Tests valid argument
        """
        self.assertEqual(validate_arg(self.VALID_ARGUMENTS[0]), None)

    def test_invalid_argument(self):
        """
        Tests invalid argument
        """
        with self.assertRaises(ValueError):
            validate_arg(self.INVALID_ARGUMENTS[0])

    def test_read_parser_args(self):
        """
        Makes sure parameters are read
        """
        params = ["ansibler.py", "--populate-platforms"]
        with self.mock_valid_args:
            with patch.object(argparse._sys, "argv", params):
                self.assertEqual(argparse._sys.argv, params)



from unittest import TestCase
from ansibler.utils.subprocesses import get_subprocess_output


class TestSubprocesses(TestCase):
    def test_get_subprocess_output(self):
        """
        Test subprocess output
        """
        out = get_subprocess_output(["echo", "Hello, world!"])
        self.assertEqual(out, "Hello, world!")

    def test_filter_subprocess_output(self):
        """
        Test grep subprocess output
        """
        out = get_subprocess_output(("echo", "-e", "Hello\nWorld"), "World")
        self.assertEqual(out, "World")



import json
import glob
from pathlib import Path
import shutil
from datetime import datetime
from typing import AnyStr, List, Optional, Tuple, Union


def create_folder_if_not_exists(path: str) -> None:
    """
    Creates dir if it does not exist.

    Args:
        path (str): dir
    """
    Path(path).mkdir(parents=True, exist_ok=True)


def create_file_if_not_exists(path: str) -> None:
    """
    Creates file if it does not exist.

    Args:
        path (str): file path
    """
    # TODO: TESTS
    Path(path).touch(exist_ok=True)


def check_folder_exists(path: str) -> bool:
    """
    Checks if a directory exists

    Args:
        path (str): dir

    Returns:
        bool: exists
    """
    return Path(path).is_dir()


def check_file_exists(path: str) -> bool:
    """
    Checks if a file exists

    Args:
        path (str): file path

    Returns:
        bool: exists
    """
    # TODO: TESTS
    return Path(path).is_file()


def list_files(
    path: str,
    pattern: Optional[str] = "*",
    absolute_path: Optional[bool] = False
) -> List[Tuple[str, datetime]]:
    """
    Lists files in a directory and returns name, datetime

    Args:
        path (str): dir
        pattern (str): only match files that satisfy this pattern
        absolute_path (bool): indicates whether returned paths should be
        absolute. When it's equal to False, only the file name is returned.

    Returns:
        (List[Tuple[str, date]]): list of file (name, date)
    """
    p = Path(path).glob(pattern)
    return [
        (
            x.name if not absolute_path else str(x.resolve()),
            datetime.fromtimestamp(x.stat().st_ctime)
        )
        for x in p
        if x.is_file()
    ]


def copy_file(
    src: str,
    destination: str,
    new_content: Optional[str] = None,
    is_json: Optional[bool] = False
) -> None:
    """
    Copies file, optionally adds new content.

    Args:
        src (str): src file
        destination (str): destination file
        new_content (str, optional): new content - defaults to None.
        is_json (bool, optional): json file? - defaults to False.

    Raises:
        shutil.SameFileError: raised when src and destination are the same file
    """
    # Create destination folder if it doesnt exist
    parent_dir = Path(destination).parents[0]
    create_folder_if_not_exists(parent_dir)

    # Copy file
    try:
        shutil.copy(src, destination)
    except shutil.SameFileError as e:
        # If content is to be overwritten, ignore SameFileError
        if new_content:
            pass
        else:
            raise e

    # Overwrite content if necessary
    if new_content:
        with open(destination, "w", encoding="utf-8") as f:
            if is_json:
                json.dump(
                    json.loads(new_content), f, ensure_ascii=False, indent=2)
            else:
                f.write(new_content)


def read_gitignore(gitignore_dir: Optional[str] = "./.gitignore") -> List[str]:
    """
    Read all files to ignore (from .gitignore)

    Args:
        gitignore_dir (str): gitignore's path. Defaults to './.gitignore'

    Returns:
        List[str]: files to ignore
    """
    files_to_ignore = []

    if not check_file_exists(gitignore_dir):
        return files_to_ignore

    for p in Path(gitignore_dir).read_text().split("\n"):
        if p and not p.startswith("#"):
            walk_gitignore(p, files_to_ignore)

    return files_to_ignore


def walk_gitignore(p: Union[Path, str], files_to_ignore: List[str]) -> None:
    """
    Reads a gitignore entry and appends it to the files_to_ignore list. If it is
    a directory, it will make sure to add all of their files and subdirs to the
    list as well.

    Args:
        p (Union[Path, str]): 
        files_to_ignore (List[str]): list of files
    """
    p_path = Path(p)

    if p_path.is_file():
        files_to_ignore.append(p_path.resolve())
    elif p_path.is_dir():
        if p.endswith("/") or p.endswith("\\"):
            files_to_ignore.extend([
                Path(to_ignore).resolve()
                for to_ignore in glob.glob(p + "*")
            ])
        else:
            files_to_ignore.extend([
                Path(to_ignore).resolve()
                for to_ignore in glob.glob(p + "/*")
            ])
    else:
        for subdir in glob.glob(p):
            walk_gitignore(subdir, files_to_ignore)



import subprocess
from typing import List, Optional


def get_subprocess_output(
    bash_cmd: List[str], filter_by: Optional[str] = None
) -> str:
    """
    Get subprocess output, filters (grep) if necessary.

    Args:
        bash_cmd (List[str]): command to run
        filter_by (str, optional): filter

    Returns:
        str: output
    """
    # Run cmd
    with subprocess.Popen(
        bash_cmd, stdout=subprocess.PIPE, encoding="UTF-8") as ps:

        # Filter
        if filter_by:
            output = subprocess.check_output(
                ("grep", filter_by), stdin=ps.stdout, encoding="UTF-8")
        else:
            output, _ = ps.communicate()

        # Wait for subprocess to terminate, then return output
        ps.wait()
        return output.strip()



def get_role_name(role_path: str, meta_file_path: str) -> str:
    """
    Gets role name from its path and meta file path.

    Args:
        role_path (str): role path
        meta_file_path (str): role meta file path.

    Returns:
        str: role name.
    """
    return meta_file_path \
        .replace("meta/main.yml", "") \
        .replace(role_path, "") \
        .strip("/").split("/")[-1]


def get_role_name_from_req_file(base_path: str, req_file_path: str) -> str:
    """
    Gets role name from its base path and requirements.yml path.

    Args:
        base_path (str): role base path
        req_file_path (str): role requirements.yml file path.

    Returns:
        str: role name.
    """
    return req_file_path \
        .replace("requirements.yml", "") \
        .replace(base_path, "") \
        .strip("/")


def get_role_full_path(role_path: str, role_name: str) -> str:
    """
    Get absolute path for a role.

    Args:
        role_path (str): role base path
        role_name (str): role name

    Returns:
        str: role absolute path
    """
    # Append / if necessary
    if role_path.endswith("/"):
        role_full_path = role_path
    else:
        role_full_path = role_path + "/"

    # Build absolute path and return
    role_full_path += role_name
    return role_full_path



from argparse import ArgumentParser, ArgumentError
from typing import Dict
from ansibler import name, __version__

ARGS = [
    {
        "arg_name": "generate-compatibility-chart",
        "arg_help": "Generates compatibilty chart",
        "arg_action": "store_true"
    },
    {
        "arg_name": "populate-platforms",
        "arg_help": "Populate platforms",
        "arg_action": "store_true",
    },
    {
        "arg_name": "role-dependencies",
        "arg_help": "Role dependencies",
        "arg_action": "store_true"
    },
    {
        "arg_name": "molecule-results-dir",
        "arg_help": "Molecule results directory " \
                    "(works for --generate-compatibility-chart only)"
    },
    {
        "arg_name": "json-file",
        "arg_help": "Overrides the JSON file used by default (ansibler.json)",
    },
    {
        "arg_name": "repository-status-template",
        "arg_help": "Repo status template used in role dependencies - can be " \
                    "a string or a valid file path"
    },
    {
        "arg_name": "variables",
        "arg_help": "Variables used to populate the repo status template. It " \
                    "must be a valid JSON file path"
    },
    {
        "arg_name": "clear-cache",
        "arg_help": "Clears ansibler cache",
        "arg_action": "store_true"
    },
    { "arg_name": "version", "arg_help": "project version" },
]


def get_user_arguments() -> Dict[str, str]:
    """
    Configures argument parser, then gets the ones entered by the user

    Returns:
        Dict[str, str]: user arguments ({ name: value, ... })
    """
    parser = configure_parser()
    user_args = read_parser_args(parser)
    return user_args


def configure_parser() -> ArgumentParser:
    """
    Instantiates the argument parser instance and adds the accepted arguments

    Returns:
        argparse.ArgumentParser: parser
    """
    parser = ArgumentParser(allow_abbrev=False)

    for arg in ARGS:
        validate_arg(arg)

        if arg.get("arg_name", None) == "version":
            parser.add_argument(
                "-v",
                f"--{arg.get('arg_name')}",
                action="version",
                version=f"{name} {__version__}",
                help=arg.get("arg_help")
            )
        elif arg.get("arg_action", None) != None:
            parser.add_argument(
                f"--{arg.get('arg_name')}",
                action=arg.get("arg_action"),
                help=arg.get("arg_help")
            )
        else:
            parser.add_argument(
                f"--{arg.get('arg_name')}",
                help=arg.get("arg_help")
            )

    return parser


def validate_arg(arg: Dict[str, str]) -> None:
    """
    Validates argument

    Args:
        arg (Dict[str, str]): argument to check

    Raises:
        ValueError: invalid argument
    """
    if not arg.get("arg_name", None) or not arg.get("arg_help", None):
        raise ValueError("Invalid argument")


def read_parser_args(parser) -> Dict[str, str]:
    """
    Reads arguments entered by the user

    Args:
        parser (argparse.ArgumentParser): parser

    Returns:
        Dict[str, str]: user arguments ({ name: value, ... })
    """
    args = parser.parse_args()
    user_args = {}

    for arg in ARGS:
        # python's argparse changes dashes (-) to underscores (_) internally
        arg_name = arg.get("arg_name")
        formatted_arg_name = arg_name.replace("-", "_")

        # Get parsed argument
        parsed_arg_val = getattr(args, formatted_arg_name, None)
        if parsed_arg_val:
            user_args[arg_name] = parsed_arg_val

    # If --version is in arguments, make sure it's the only one
    if "version" in user_args.keys() and len(user_args) > 1:
        raise ArgumentError("Invalid option: \"version\"")

    return user_args



import os
import pathlib
import re
import json
import asyncio
from json.decoder import JSONDecodeError
from typing import Any, Coroutine, Dict, List, Optional
from ruamel.yaml import YAML
from ansibler.utils.subprocesses import get_subprocess_output
from ansibler.role_dependencies.role_info import get_role_name_from_req_file
from ansibler.role_dependencies.galaxy import get_from_ansible_galaxy
from ansibler.exceptions.ansibler import CommandNotFound, MetaYMLError, RolesParseError
from ansibler.role_dependencies.cache import (
    read_roles_metadata_from_cache, cache_roles_metadata, append_role_to_cache
)
from ansibler.utils.files import (
    check_folder_exists,
    check_file_exists,
    list_files,
    copy_file,
    check_file_exists,
    create_file_if_not_exists,
    read_gitignore
)


ROLES_PATTERN = r"\[.*\]"


async def generate_role_dependency_chart(
    json_file: Optional[str] = "./ansibler.json",
    template: Optional[str] = None,
    variables: Optional[str] = None
) -> Coroutine[None, None, None]:
    """
    Generates role dependency charts. Uses caches whenever possible.
    """
    # TODO: TESTS
    # Read role paths
    role_paths = parse_default_roles(get_default_roles())
    is_playbook = check_file_exists("./ansible.cfg")

    # Read cache
    cache = read_roles_metadata_from_cache()

    # Generate cache if necessary
    if cache is None:
        cache = cache_roles_metadata(role_paths)

    # Task pool
    tasks = []

    paths = [os.path.abspath("./")] if not is_playbook else role_paths
    for role_path in paths:
        if not check_folder_exists(role_path):
            continue

        # Read gitignore
        files_to_ignore = read_gitignore(
            pathlib.Path(role_path) / pathlib.Path(".gitignore"))

        # List ansible dirs
        if is_playbook:
            files = list_files(role_path, "**/meta/main.yml", True)
        else:
            files = list_files(role_path, "meta/main.yml", True)
            role_path = "/".join(role_path.split("/")[:-1])

        for f in files:
            # Skip if in gitignore
            if pathlib.Path(f[0]) in files_to_ignore:
                continue

            # Make sure we're dealing with an ansible project (role or playbook)
            if not is_ansible_dir(f[0].replace("meta/main.yml", "")):
                continue

            # Get the role name
            req_file = f[0].replace("meta/main.yml", "requirements.yml")
            if is_playbook:
                role_name = get_role_name_from_req_file(role_path, req_file)
            else:
                role_name = role_path.split("/")[-1]

            # Append task to the pool
            tasks.append(
                asyncio.ensure_future(
                    generate_single_role_dependency_chart(
                        role_name,
                        req_file,
                        role_path,
                        cache,
                        os.path.basename(json_file),
                        role_paths,
                        template=template,
                        variables=variables
                    )
                )
            )

    # Execute tasks
    await asyncio.gather(*tasks)
    print("Done")


def get_default_roles() -> str:
    """
    Get raw DEFAULT_ROLES_PATH from running ansible-config dump

    Raises:
        CommandNotFound: raised when command not available

    Returns:
        str: command output
    """
    # Get default roles
    bash_cmd = ["ansible-config", "dump"]
    default_roles = get_subprocess_output(bash_cmd, "DEFAULT_ROLES_PATH")

    # Check if valid
    if not default_roles or "DEFAULT_ROLES_PATH" not in default_roles:
        raise CommandNotFound("Could not run", " ".join(bash_cmd))

    return default_roles


def parse_default_roles(default_roles: str) -> List[str]:
    """
    Parses default roles from raw command output

    Args:
        default_roles (str): raw roles dump, straight from cmd output

    Raises:
        RolesParseError: default_roles doesnt have the expected format

    Returns:
        List[str]: list of role paths
    """
    # Find list of roles
    match = re.search(ROLES_PATTERN, default_roles)
    if not match:
        raise RolesParseError(f"Couldn't parse roles from: {default_roles}")

    # Parse them
    roles = match.group(0).strip("[").strip("]").replace("'", "").split(",")
    return [role.strip() for role in roles]


def is_ansible_dir(directory: str) -> bool:
    """
    Checks if dir is an ansible playbook or role.

    Args:
        directory (str): dir to check

    Returns:
        bool: whether an ansible playbook or role
    """
    return any((
        check_file_exists(directory + "meta/main.yml"),
        check_folder_exists(directory + "molecule/")
    ))


async def generate_single_role_dependency_chart(
    role_name: str,
    requirement_file: str,
    role_base_path: str,
    cache: Dict[str, Any],
    json_file: Optional[str] = "ansibler.json",
    role_paths: Optional[str] = [],
    template: Optional[str] = None,
    variables: Optional[str] = None
) -> Coroutine[None, None, None]:
    # TODO: TESTS
    try:
        json_basename = os.path.basename(json_file)
        await role_dependency_chart(
            requirement_file,
            role_base_path,
            cache,
            json_file=json_basename,
            role_paths=role_paths,
            template=template,
            variables=variables
        )
    except (ValueError, MetaYMLError) as e:
        print(
            f"\tCouldnt generate dependency chart for {role_name}: {e}")


async def role_dependency_chart(
    requirement_file: str,
    role_base_path: str,
    cache: Dict[str, Any],
    json_file: Optional[str] = "ansibler.json",
    role_paths: Optional[str] = [],
    template: Optional[str] = None,
    variables: Optional[str] = None
) -> Coroutine[None, None, None]:
    # TODO: TESTS
    # Get role's name
    role_name = get_role_name_from_req_file(role_base_path, requirement_file)

    print(f"Generating role dependency for {role_name}")

    role_dependencies = []

    # Read dependencies
    dependencies = read_dependencies(requirement_file)
    # If there's at least one dependency, add headers
    if len(dependencies):
        role_dependencies.append([
            "Dependency",
            "Description",
            "Supported OSes",
            "Status"
        ])
    else:
        print(f"\tNo dependencies found in {role_name}")

    for dep in dependencies:
        if dep is None:
            print(f"\tFound invalid dependency in {role_name}")
            continue

        dep_name = dep.split(".")[-1]
        print(f"\tReading dependency {dep}")
        dependency_metadata = cache.get(dep_name, {})

        # if not found locally, try getting from ansible-galaxy
        if not dependency_metadata:
            if role_paths:
                print(f"\tDoing full re-scan...")
                new_cache = cache_roles_metadata(role_paths, cache)
                return await role_dependency_chart(
                    requirement_file,
                    role_base_path,
                    new_cache,
                    json_file=json_file
                )

            print(f"\tReading dependency {dep} from ansible-galaxy")
            dependency_metadata = get_from_ansible_galaxy(dep)
            append_role_to_cache(dep_name, dependency_metadata, cache)

        role_dependencies.append(
            get_dependency_metadata(
                dependency_metadata,
                role_base_path \
                    if role_base_path.endswith("/") \
                    else role_base_path + "/" + role_name,
                template,
                variables
            )
        )

    if role_base_path.startswith("./"):
        role_path = "/" + role_base_path + "/" + role_name + "/"
    else:
        role_path = role_base_path + "/" + role_name + "/"

    data = {}
    ansibler_json_file = role_path + json_file

    if not check_file_exists(ansibler_json_file):
        create_file_if_not_exists(ansibler_json_file)

    try:
        with open(ansibler_json_file) as f:
            data = json.load(f)

        if isinstance(data, list):
            raise JSONDecodeError()
    except (JSONDecodeError, FileNotFoundError):
        data = {}

    data["role_dependencies"] = role_dependencies

    copy_file(ansibler_json_file, ansibler_json_file, json.dumps(data), True)
    print(f"\tGenerated role dependency chart for {role_name}")


def read_dependencies(requirements_file_path: str) -> List[str]:
    """
    Reads a role dependencies from requirements.yml

    Args:
        requirements_file_path (str): requirements.yml path

    Returns:
        List[str]: list of dependency names
    """
    # TODO: TESTS
    data = {}
    try:
        with open(requirements_file_path) as f:
            yaml = YAML()
            data = yaml.load(f)
    except FileNotFoundError:
        return []

    if data is None:
        return []

    return [role["name"] for role in data.get("roles", []) if "name" in role]


def get_dependency_metadata(
    dependency_metadata: Dict[str, Any],
    role_base_path: str,
    template: Optional[str] = None,
    variables: Optional[str] = None
) -> List[str]:
    """
    Returns formatted dependency's metadata

    Args:
        dependency_metadata (Dict[str, Any]): metadata
        role_base_path (str): role path
        template (str): repo status template
        variables (Dict[str, str]): role variables

    Returns:
        List[str]: formatted metadata
    """
    # TODO: TESTS
    supported = get_role_dependency_supported_oses(dependency_metadata)
    supported = f"<center>{supported}</center>" if supported else supported

    if template:
        status = get_role_dependency_status_from_template(template, variables)
    else:
        status = get_role_dependency_status(dependency_metadata, role_base_path)

    status = f"<center>{status}</center>" if status else status

    return [
        get_role_dependency_link(dependency_metadata),
        get_role_dependency_description(dependency_metadata),
        supported,
        status
    ]


def get_role_dependency_link(metadata: Dict[str, Any]) -> str:
    """
    Returns role dependency link

    Args:
        metadata (Dict[str, Any]): role metadata

    Returns:
        str: role dependency link
    """
    role_name = metadata.get("role_name", None)
    namespace = metadata.get("namespace", None)

    if not namespace or not role_name:
        raise ValueError(
            f"Can not generate dependency link for {namespace}.{role_name}")
    
    return f"<b>" \
           f"<a href=\"https://galaxy.ansible.com/{namespace}/{role_name}\" " \
           f"title=\"{namespace}.{role_name} on Ansible Galaxy\" target=\"_" \
           f"blank\">{namespace}.{role_name}</a></b>"


def get_role_dependency_description(metadata: Dict[str, Any]) -> str:
    """
    Returns role dependency description.

    Args:
        metadata (Dict[str, Any]): role metadata

    Returns:
        str: description
    """
    description = metadata.get("description")

    if not description:
        f"Can not get description for {metadata.get('role_name', 'role')}"

    return description


def get_role_dependency_supported_oses(metadata: Dict[str, Any]) -> str:
    """
    Returns list of supported OSes for a specific role

    Args:
        metadata (Dict[str, Any]): role metadata

    Returns:
        str: [description]
    """
    platforms = metadata.get("platforms", [])
    repository = metadata.get("repository", None)

    supported_oses = []
    for platform in platforms:
        name = str(platform.get("name", None)).lower()

        img = "https://gitlab.com/megabyte-labs/assets/-/raw/master/icon/"
        alt = ""
        if "arch" in name:
            img += "archlinux.png"
            alt = "Arch"
        elif "centos" in name or "el" in name:
            img += "centos.png"
            alt = "EL"
        elif "debian" in name:
            img += "debian.png"
            alt = "Debian"
        elif "fedora" in name:
            img += "fedora.png"
            alt = "Fedora"
        elif "freebsd" in name:
            img += "freebsd.png"
            alt = "FreeBSD"
        elif "mac" in name:
            img += "macos.png"
            alt = "MacOS"
        elif "ubuntu" in name:
            img += "ubuntu.png"
            alt = "Ubuntu"
        elif "windows" in name:
            img += "windows.png"
            alt = "Windows"
        elif "generic" in name:
            img += "linux.png"
            alt = "GenericUNIX"
        else:
            raise ValueError(f"Could not find icon for platform {name}")

        if repository:
            supported_oses.append(
                f"<img src=\"{img}\" href=\"{repository}#supported-operating" \
                f"-systems\" alt=\"{alt}\" />")
        else:
            supported_oses.append(
                f"<img src=\"{img}\" alt=\"{alt}\" />")

    supported_oses = "".join(supported_oses)
    return supported_oses if supported_oses else "❔"


def get_role_dependency_status(metadata: Dict[str, Any], role_path: str) -> str:
    """
    Returns role status

    Args:
        metadata (Dict[str, Any]): role metadata
        role_path (str): role path

    Returns:
        str: role status
    """
    # Looks for .variables.json
    role_name = metadata.get("role_name", None)
    if not role_path.endswith("/"):
        variables_json = role_path + "/" + "variables.json"
    else:
        variables_json = role_path + "variables.json"

    # Return question mark if it the file does not exist
    if not role_name or not check_file_exists(variables_json):
        return "❔"

    # Read variables json file
    data = {}
    try:
        with open(variables_json) as f:
            data = json.load(f)
    except:
        pass

    status = data.get("role_dependencies_status_format", None)
    if status is None:
        return "❔"

    # Replace {{ role_name }} ocurrences
    status = status.replace(r"{{ role_name }}", role_name)
    return status


def get_role_dependency_status_from_template(
    template: str, variables: Dict[str, str]
) -> str:
    """
    Returns role status from template

    Args:
        template (str): repo status template
        variables (Dict[str, str]): role variables

    Returns:
        str: role status
    """
    new_template = template[:]

    for variable, value in variables.items():
        new_template = new_template.replace("{{ " + variable + " }}", value)
        new_template = new_template.replace("{{" + variable + "}}", value)

    return new_template



from typing import Any, Dict
from ansibler.utils.subprocesses import get_subprocess_output
from ansibler.exceptions.ansibler import RoleMetadataError


def get_from_ansible_galaxy(role: str) -> Dict[str, Any]:
    """
    Gets role metadata from ansible-galaxy.

    Args:
        role (str): role in the form {{role_namespace}}.{{role_name}}

    Raises:
        RoleMetadataError: raised when role not found

    Returns:
        Dict[str, Any]: role metadata
    """
    # Run ansible-galaxy info {{ role }}
    out = get_subprocess_output(["ansible-galaxy", "info", role], "description")

    # Search for description, raise error if role not found
    if "description" in out:
        description = out.replace("description:", "").strip()
        if not description:
            print(f"WARNING: Role {role} not found")
            raise RoleMetadataError("Role not found")
    else:
        raise RoleMetadataError("Role not found")

    # Extract data
    role_data = role.split(".")
    return {
        "namespace": role_data[0],
        "role_name": role_data[1],
        "description": description,
        "repository": None,
        "repository_status": None
    }



class BaseAnsiblerException(Exception):
    message = "Error"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args)
        self.__class__.message = kwargs.get("message", self.message)

    def __str__(self) -> str:
        return self.__class__.message


class CommandNotFound(BaseAnsiblerException):
    message = "Command not found"


class RolesParseError(BaseAnsiblerException):
    message = "Could not parse default roles"


class MetaYMLError(BaseAnsiblerException):
    message = "Invalid meta/main.yml"


class RoleMetadataError(BaseAnsiblerException):
    message = "Role metadata error"


class MoleculeTestsNotFound(BaseAnsiblerException):
    message = "Molecule tests not foound"


class MoleculeTestParseError(BaseAnsiblerException):
    message = "Could not parse molecule test file"


class NoPackageJsonError(BaseAnsiblerException):
    message = "No package.json"



from setuptools import setup, find_packages
from ansibler import __version__, name

NAME = name
REPO = name

VERSION = __version__
ARCHIVE = f"v_{'_'.join(VERSION.split('.'))}.tar.gz"

setup(
    name=NAME,
    packages=find_packages(),
    version=VERSION,
    license="MIT",
    description="Generate JSON data that describes the dependencies of an " \
        "Ansible playbook/role. Also, automatically generate OS compatibil" \
        "ity charts using Molecule.",
    author="Renny Montero",
    author_email="rennym19@gmail.com",
    url=f"https://gitlab.com/megabyte-labs/python/{REPO}/",
    download_url=f"https://gitlab.com/megabyte-labs/python/{REPO}/archive/" \
        f"{ARCHIVE}",
    keywords=["ANSIBLE", "DEPENDENCY", "ROLE", "MOLECULE", "CHARTS", "TEST"],
    install_requires=[
        "ruamel.yaml"
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [f"{NAME} = ansibler.run:main"]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import json
import math
from typing import Any, Dict, List, Optional, Union
from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError
from ansibler.utils.files import create_folder_if_not_exists


def populate_platforms(json_file: Optional[str] = "./ansibler.json") -> None:
    # TODO: TESTS
    # Read from package.json
    data = read_json_file(json_file)
    compatibility = data.get("compatibility_matrix", [])
    compatibility = [] if len(compatibility) <= 1 else compatibility[1:]

    # Generate platforms value
    platforms, supported, unsupported = [], [], []
    for platform in compatibility:
        os = f"{platform[0]}-{platform[1]}"
        if "✅" in platform[2]:
            version = get_formatted_os_version(platform[0], platform[1])

            platforms.append({
                "name": platform[0],
                "versions": [version]
            })

            supported.append(os)
        else:
            unsupported.append(os)

    # Populate meta/main.yml
    meta_main = read_meta_main("./meta/main.yml")
    galaxy_info = meta_main.get("galaxy_info", {})

    old_platforms = galaxy_info.get("platforms", [])
    platforms = merge_platforms(
        platforms, old_platforms, supported, unsupported)
    platforms = join_platforms(platforms)

    galaxy_info["platforms"] = platforms if platforms else None
    meta_main["galaxy_info"] = galaxy_info

    # Save
    create_folder_if_not_exists("./meta/")
    out = "./meta/main.yml"
    with open(out, "w") as f:
        yaml = YAML()
        yaml.explicit_start = True
        yaml.dump(meta_main, f)

    print("Done")


def read_json_file(json_file_path: str) -> Dict[str, Any]:
    # TODO: TESTS
    with open(json_file_path) as f:
        return json.load(f)


def read_meta_main(meta_main_path: str) -> Dict[str, Any]:
    # TODO: TESTS
    try:
        with open(meta_main_path) as f:
            yaml = YAML()
            return yaml.load(f)
    except (FileNotFoundError, YAMLError):
        return {}


def merge_platforms(
    current_platforms: List[Dict[str, Any]],
    old_platforms: List[Dict[str, Any]],
    supported: List[str],
    unsupported: List[str]
) -> List[Dict[str, Any]]:
    """
    Merge platforms from package.json's blueprint.compatibility and
    meta/main.yml. Only removes a platform when / if it's marked as unsuccessful
    in the blueprint.compatibility field.

    Args:
        current_platforms (List[Dict[str, Any]]): current platforms data
        old_platforms (List[Dict[str, Any]]): old platforms data
        supported (List[Dict[str, Any]]): supported OSes ({name}-{version})
        unsupported (List[Dict[str, Any]]): unsupported OSes ({name}-{version})

    Returns:
        List[Dict[str, Any]]: merged platforms
    """
    # TODO: TESTS
    res = current_platforms[:]

    if old_platforms is None:
        old_platforms = []

    for old_platform in old_platforms:
        name = old_platform.get("name", None)
        versions = old_platform.get("versions", [])

        if name is None:
            continue

        for version in versions:
            os = f"{name}-{get_formatted_os_version(name, version)}"
            added = False

            if os not in unsupported and os not in supported:
                for current_platform in res:
                    cur_name = current_platform.get("name")
                    if cur_name == name:
                        cur_versions = current_platform.get("versions", [])
                        cur_versions.append(
                            get_formatted_os_version(name, version))
                        supported.append(os)
                        added = True

                if not added:
                    res.append({
                        "name": name,
                        "versions": [get_formatted_os_version(name, version)]
                    })
                    supported.append(os)

    return res


def join_platforms(platforms: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Joins platform versions.

    Args:
        platforms (List[Dict[str, Any]]): platforms

    Returns:
        List[Dict[str, Any]]: joined platforms
    """
    # TODO: TESTS
    res = []
    added_names = []

    for platform in platforms:
        name = platform.get("name")
        versions = platform.get("versions", [])

        if name not in added_names:
            res.append({
                "name": name.split(" ")[0],
                "versions": [
                    get_formatted_os_version(name, v)
                    for v in versions
                ]
            })
            added_names.append(name.split(" ")[0])
        else:
            for p in res:
                added_platform_name = p.get("name")
                added_versions = p.get("versions", [])
                if added_platform_name == name:
                    for version in versions:
                        if version not in added_versions:
                            added_versions.append(
                                get_formatted_os_version(name, version))
                            p["versions"] = sorted(
                                added_versions, key=lambda x: str(x))

    return res


def get_formatted_os_version(os: str, version: str) -> Union[float, int, str]:
    # TODO TESTS
    if os.lower() == "windows":
        return "all"
    elif isinstance(version, str) and version.replace(".", "", 1).isnumeric():
        if "." in version:
            if int(version.split(".")[-1]) == 0:
                return int(version.split(".")[0])
            return float(version)
        else:
            return int(version)
    elif isinstance(version, str) and \
        not version.replace(".", "", 1).isnumeric():
        if os.lower() == "ubuntu" and "(" in version:
            print(os, version)
            return version.split("(")[-1].split(" ")[0].lower()

        return version
    else:
        version_ceil = math.ceil(float(version))
        version_float = float(version)

        if version_ceil == version_float:
            return version_ceil

        return version



from json.decoder import JSONDecodeError
import os
from datetime import datetime
import re
import json
from typing import Any, Dict, List, Optional, Tuple
from ansibler.utils.files import check_folder_exists, list_files
from ansibler.exceptions.ansibler import (
    MoleculeTestParseError, MoleculeTestsNotFound
)
from ansibler.molecule_test.parse import parse_test


MOLECULE_RESULTS_DIR = "./molecule-results/"
FILTER_FILES_PATTERN = r"\d{4}-\d{2}-\d{2}-.*.txt"


def generate_compatibility_chart(
    molecule_results_dir: Optional[str] = None,
    json_file: Optional[str] = "./ansibler.json"
) -> None:
    if molecule_results_dir is None:
        molecule_results_dir = MOLECULE_RESULTS_DIR

    # TODO: TESTS
    # Check molecule-results dir exists
    if not check_folder_exists(molecule_results_dir):
        raise MoleculeTestsNotFound("Couldn't find molecule results dir")

    # Get list of molecule test files
    test_files = list_files(molecule_results_dir, absolute_path=True)
    test_files = [
        (file_name, get_test_file_date(file_name))
        for file_name, _ in test_files
        if re.search(FILTER_FILES_PATTERN, file_name)
    ]

    # Prepare to build blueprint.compatibility array
    compat = [["OS Family", "OS Version", "Status", "Idempotent", "Tested On"]]
    temp_compat = {}

    # Parse test files
    for test_file, test_date in test_files:
        try:
            converge, idempotence = read_molecule_tests(test_file)

            # Skip if converge is invalid
            if not converge:
                continue

            # Read play recaps and add them to temp_compat if they are the most
            # recent for a given OS
            play_recap = converge.get("play_recap", [])
            for recap in play_recap:
                os, recap_summary = get_play_recap_summary(
                    recap, idempotence, test_date)

                if os in temp_compat and test_date < temp_compat[os]["added"]:
                    continue

                temp_compat[os] = recap_summary
        except MoleculeTestParseError as e:
            print(f"Error while parsing molecule test file {test_file}: {e}")

    # Add to blueprint.compatibility
    add_items_to_blueprint_compatibility(temp_compat, compat)

    # Populate output file
    data = {}
    try:
        with open(json_file) as f:
            data = json.load(f)

        if isinstance(data, list):
            raise JSONDecodeError()
    except (JSONDecodeError, FileNotFoundError):
        data = {}

    data["compatibility_matrix"] = compat

    # Save
    with open(json_file, "w") as f:
        json.dump(data, f, ensure_ascii=False, indent=2)

    print("Done")


def get_test_file_date(test_file_name) -> datetime:
    """
    Extracts test date from its name.
    Must be in the following format: YEAR-MONTH-DAY-etc.txt

    Args:
        test_file_name (str): test date
    """
    # TODO: TESTS
    basename = os.path.basename(test_file_name)
    return datetime.fromisoformat(basename[:10])


def read_molecule_tests(
    test_file: str
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Reads converge and idempotence tests from a test file.

    Args:
        test_file (str): test file path

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any]]: converge and idempotence tests
    """
    # TODO: TESTS
    molecule_test_dump = ""

    with open(test_file) as f:
        molecule_test_dump = f.read()

    test = parse_test(molecule_test_dump)
    converge = test.get("converge", {})
    idempotence = test.get("idempotence", {})

    return converge, idempotence


def get_play_recap_summary(
    recap: Dict[str, Any], idempotence: Dict[str, Any], test_date: datetime
) -> Tuple[str, Dict[str, Any]]:
    """
    Returns a summary of a converge play recap.

    Args:
        recap (Dict[str, Any]): play recap
        idempotence (Dict[str, Any]): corresponding idempotence test

    Returns:
        Tuple[str, Dict[str, Any]]: os, play recap summary
    """
    # TODO: TESTS
    os_name = recap.get("os_name")
    os_version = recap.get("os_version")
    success = did_play_succeed(recap)
    idempotent = is_idempotent(recap, idempotence)

    os = f"{os_name}-{os_version}"

    return os, {
        "os_family": os_name,
        "os_version": os_version,
        "success": success,
        "idempotent": idempotent,
        "added": test_date
    }


def did_play_succeed(
    recap: Dict[str, Any],
    idempotency_play: Optional[bool] = False
) -> bool:
    """
    Checks if a play succeeded.

    Args:
        recap (Dict[str, Any]): play recap
        idempotency_play (bool, optional): idempotence play? Defaults to False.

    Returns:
        bool: whether successful or not
    """
    # TODO: TESTS
    ok = recap.get("ok", 0)
    failed = recap.get("failed", 0)
    unreachable = recap.get("unreachable", 0)

    if idempotency_play:
        changed = recap.get("changed", 0)
        return ok > 0 and not failed and not unreachable and not changed

    return ok > 0 and failed == 0 and unreachable == 0


def is_idempotent(
    recap: Dict[str, Any], idempotence_test: Dict[str, Any]
) -> bool:
    """
    Checks if a test was idempotent.

    Args:
        recap (Dict[str, Any]): play recap
        idempotence_test (Dict[str, Any]): corresponding idempotence test

    Returns:
        bool: whether idempotent or not
    """
    # TODO: TESTS
    idempotence_results = idempotence_test.get("play_recap", [])

    if not idempotence_results:
        return None

    recap_os_name =  recap.get("os_name")
    recap_os_version = recap.get("os_version")

    for result in idempotence_results:
        idp_os_name = result.get("os_name")
        idp_os_version = result.get("os_version")

        if idp_os_name == recap_os_name and idp_os_version == recap_os_version:
            return did_play_succeed(result, idempotency_play=True)

    return None


def add_items_to_blueprint_compatibility(
    items: List[Dict[str, Any]], compat: List[List[str]]
) -> None:
    """
    Appends items to the final blueprint compatibility array.

    Args:
        items (List[Dict[str, Any]]): items to add
        compat (List[List[str]]): List to append the items to.
    """
    # TODO: TESTS
    for _, data in items.items():
        idempotent = data.get("idempotent", None)
        if idempotent == True:
            idempotent = "✅"
        else:
            idempotent = "❌"

        compat.append([
            data["os_family"],
            data["os_version"],
            f"<center>{'✅' if data['success'] else '❌'}</center>",
            f"<center>{idempotent}</center>",
            custom_strftime("%B {S}, %Y", data["added"])
        ])


def custom_strftime(format: str, t: datetime) -> str:
    """
    Custom time format containing English day suffixes (st, nd, rd, th).

    Args:
        format (str): format
        t (datetime): datetime

    Returns:
        (str): formatted datetime
    """
    return t.strftime(format).replace('{S}', str(t.day) + suffix(t.day))


def suffix(d: int) -> str:
    """
    Retuns DAY NUMBER date suffix.

    Args:
        d (int): day

    Returns:
        str: suffix
    """
    # TODO: TESTS
    if 11 <= d <= 13:
        return "th"
    else:
        return { 1: "st", 2: "nd", 3: "rd"}.get(d % 10, "th")



import json
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, Dict, List, Optional
from ruamel.yaml import YAML
from ansibler.utils.files import create_folder_if_not_exists, list_files
from ansibler.exceptions.ansibler import MetaYMLError
from ansibler.utils.files import create_folder_if_not_exists
from ansibler.role_dependencies.role_info import get_role_name


META_FILES_PATTERN = "**/meta/main.yml"
CACHE_MAP_DIR = str(Path.home()) + "/.local/megabytelabs/ansibler/"
CACHE_MAP_FILE = "role_metadata"


def read_roles_metadata_from_cache() -> Dict[str, Any]:
    # TODO: TESTS
    try:
        cache = None
        with open(CACHE_MAP_DIR + CACHE_MAP_FILE) as f:
            cache = json.load(f)
        if not cache:
            return None
        print(f"Read cache from {CACHE_MAP_DIR}{CACHE_MAP_FILE}")
    except (FileNotFoundError, JSONDecodeError):
        pass

    return cache


def cache_roles_metadata(
    roles_path: List[str],
    current_cache: Optional[Dict[str, Any]] = {}
) -> Dict[str, Any]:
    # TODO: TESTS
    # Create cache folder if it does not exist
    create_folder_if_not_exists(CACHE_MAP_DIR)

    cache = {**current_cache}

    for role_path in roles_path:
        meta_files = list_files(role_path, META_FILES_PATTERN, True)

        for meta_file in meta_files:
            meta_file_path = meta_file[0]
            role_name = get_role_name(role_path, meta_file_path)

            try:
                cache_single_role_metadata(
                    meta_file_path, role_path, role_name, cache)
            except:
                pass

    with open(CACHE_MAP_DIR + CACHE_MAP_FILE, "w", encoding="utf-8") as f:
        json.dump(cache, f, ensure_ascii=False, indent=2)

    if not current_cache:
        print("Role metadata cached")

    return cache


def cache_single_role_metadata(
    meta_file_path: str, role_path: str, role_name: str, cache: Dict[str, Any]
) -> None:
    # TODO: TESTS
    # Read and parse meta/main.yml
    data = {}
    with open(meta_file_path) as f:
        yaml = YAML()
        data = yaml.load(f)

    err_msg = f"Invalid meta/main.yml in: {role_path}"
    if not role_path.endswith("/"):
        err_msg += "/"
    err_msg += role_name

    if data is None:
        print(err_msg)
        raise MetaYMLError(err_msg)

    # Read galaxy_info
    galaxy_info = data.get("galaxy_info", {})

    # Raise exceptions if invalid meta/main.yml
    if not galaxy_info:
        print(err_msg)
        raise MetaYMLError(err_msg)

    if "role_name" not in galaxy_info or \
        "author" not in galaxy_info or \
        "description" not in galaxy_info:
        print(err_msg)
        raise MetaYMLError(err_msg)

    # Build cache map for this single role
    metadata = {
        "role_name": galaxy_info.get("role_name"),
        "namespace": galaxy_info.get("author"),
        "description": galaxy_info.get("description"),
        "platforms": galaxy_info.get("platforms", []),
        "repository": galaxy_info.get("repository", None),
        "repository_status": galaxy_info.get("repository_status", None)
    }

    # Append to cache
    cache[role_name] = metadata


def append_role_to_cache(
    role_name: str, metadata: str, cache: Dict[str, Any]
) -> None:
    # TODO: TESTS
    # Append to cache
    cache[role_name] = metadata

    # Rewrite cache
    with open(CACHE_MAP_DIR + CACHE_MAP_FILE, "w", encoding="utf-8") as f:
        json.dump(cache, f, ensure_ascii=False, indent=2)


def clear_cache() -> None:
    """
    Clears ansibler cache
    """
    # TODO: TESTS
    Path(CACHE_MAP_DIR + CACHE_MAP_FILE).unlink(missing_ok=True)



import re
from typing import Any, Dict, List, Tuple
from ansibler.exceptions.ansibler import MoleculeTestParseError


CONVERGE_START_PATTERN = r"INFO(\s)+Running(.*)converge"
IDEMPOTENCE_START_PATTERN = r"INFO(\s)+Running(.*)idempotence"
PLAY_FINISH_PATTERN = r"INFO(\s)+(.*)"
PLAY_NAME_PATTERN = r"INFO\s+Running.*>\s*(\w.*)"
PLAY_RECAP_PATTERN = r"(PLAY RECAP.+?\s)((.+?\s)*)"

PLAY_RECAP_OS_NAME_PATTERN = r"^\s?[^\s]*"
OK_COUNT_PATTERN = r"ok=(\d*)"
CHANGED_COUNT_PATTERN = r"changed=(\d*)"
UNREACHABLE_COUNT_PATTERN = r"unreachable=(\d*)"
FAILED_COUNT_PATTERN = r"failed=(\d*)"
SKIPPED_COUNT_PATTERN = r"skipped=(\d*)"
RESCUED_COUNT_PATTERN = r"rescued=(\d*)"
IGNORED_COUNT_PATTERN = r"ignored=(\d*)"


def parse_test(molecule_test_dump: str) -> Dict[str, Any]:
    """
    Parses a molecule test - includes 'converge' and 'idempotence'

    Args:
        molecule_test_dump (str): molecule test dump

    Returns:
        Dict[str, Any]: parsed molecule test
    """
    test = {}

    converge_test = parse_play(CONVERGE_START_PATTERN, molecule_test_dump)
    if converge_test:
        test["converge"] = converge_test

    try:
        idempotence_test = parse_play(
            IDEMPOTENCE_START_PATTERN, molecule_test_dump)
    except MoleculeTestParseError:
        idempotence_test = None

    if idempotence_test:
        test["idempotence"] = idempotence_test

    return test


def parse_play(play_pattern: str, molecule_test_dump: str) -> Dict[str, Any]:
    """
    Extracts a PLAY from a molecule test dump

    Args:
        play_pattern (str): play start pattern (regexp)
        molecule_test_dump (str): molecule test dump

    Raises:
        MoleculeTestParseError: raised when PLAY not found

    Returns:
        Dict[str, Any]: formatted play
    """
    play = {}

    m = re.search(play_pattern, molecule_test_dump)
    if not m:
        raise MoleculeTestParseError("PLAY not found")

    # Get start and end index
    play_info_dump = m.group()
    play_start, play_end = m.span()[0], -1

    m = re.search(
        PLAY_FINISH_PATTERN,
        molecule_test_dump[play_start + len(play_info_dump):]
    )
    if m:
        play_end = len(molecule_test_dump[:play_start]) + m.span()[1]

    # Parse play name and play recap info
    play_dump = molecule_test_dump[play_start:play_end]
    play["play_name"] = parse_play_name(play_info_dump)
    play["play_recap"] = parse_play_recap(play_dump)

    return play 


def parse_play_name(play_dump: str) -> str:
    """
    Extracts the name of the PLAY from the dump.

    Args:
        play_dump (str): molecule test dump

    Raises:
        MoleculeTestParseError: raised when the play name wasn't found

    Returns:
        str: play name
    """
    m = re.search(PLAY_NAME_PATTERN, play_dump)
    if not m:
        raise MoleculeTestParseError("No PLAY NAME found")

    return m.group(1)


def parse_play_recap(play_dump: str) -> List[Dict[str, Any]]:
    """
    Parses PLAY RECAP

    Args:
        play_dump (str): molecule test dump (play section)

    Returns:
        List[Dict[str, Any]]: list of recaps per OS
    """
    recap = []
    play_recap_dump = parse_play_recap_dump(play_dump)

    recap_lines = play_recap_dump.splitlines()
    for recap_line in recap_lines:
        os_name, os_version = parse_os(recap_line)
        if not os_name:
            continue

        recap.append({
            "os_name": os_name,
            "os_version": os_version,
            "ok": parse_recap_value(
                OK_COUNT_PATTERN, recap_line),
            "changed": parse_recap_value(
                CHANGED_COUNT_PATTERN, recap_line),
            "unreachable": parse_recap_value(
                UNREACHABLE_COUNT_PATTERN, recap_line),
            "failed": parse_recap_value(
                FAILED_COUNT_PATTERN, recap_line),
            "skipped": parse_recap_value(
                SKIPPED_COUNT_PATTERN, recap_line),
            "rescued": parse_recap_value(
                RESCUED_COUNT_PATTERN, recap_line),
            "ignored": parse_recap_value(
                IGNORED_COUNT_PATTERN, recap_line)
        })

    return recap


def parse_play_recap_dump(play_dump: str) -> str:
    """
    Parses PLAY RECAP dump, without any formatting

    Args:
        play_dump (str): molecule test PLAY dump

    Raises:
        MoleculeTestParseError: raised when PLAY RECAP is not found

    Returns:
        str: play recap dump
    """
    m = re.search(PLAY_RECAP_PATTERN, play_dump)
    if not m:
        raise MoleculeTestParseError("Could not parse PLAY RECAP")
    return m.group(2).strip()


def parse_os(recap: str) -> Tuple[str, str]:
    """
    Parses OS name and version from a PLAY RECAP line

    Args:
        recap (str): play recap line

    Returns:
        Tuple[str, str]: os name, version
    """
    m = re.search(PLAY_RECAP_OS_NAME_PATTERN, recap)
    if not m:
        return None, None

    # Split to get name, version
    os_data = m.group().split("-")

    if len(os_data) > 1:
        os_name = os_data[0]

        version = None
        version_index = None

        os_vers = os_data[1:]
        for i, v in enumerate(os_vers):
            if v.replace(".", "").isnumeric():
                version = v
                version_index = i
                break

        if version is None:
            version_index = 0
            version = os_vers[0]

        codename = version if len(os_vers) <= 1 else f"{version} ("
        remaining_text = []
        for i, c in enumerate(os_vers):
            if i != version_index:
                remaining_text.append(c)

        if len(remaining_text) >= 1:
            if "centos" in os_name.lower():
                os_name += f" {' '.join(remaining_text).title()}"
                codename = codename.rstrip(" (")
            else:
                codename += f"{' '.join(remaining_text).title()})"

        return os_name, codename

    return os_data[0], None


def parse_recap_value(pattern: str, recap: str) -> int:
    """
    Parses a PLAY RECAP value (ok count, failed count, etc..)

    Args:
        pattern (str): pattern to use
        recap (str): play recap line

    Returns:
        int: parsed value
    """
    m = re.search(pattern, recap)
    if not m:
        return -1
    return int(m.group(1))



import asyncio
import sys
from ansibler.args.cmd import get_user_arguments
from ansibler.compatibility.chart import generate_compatibility_chart
from ansibler.platforms.populate import populate_platforms, read_json_file
from ansibler.role_dependencies.dependencies import (
    generate_role_dependency_chart
)
from ansibler.role_dependencies.cache import clear_cache
from ansibler.utils.files import check_file_exists
from ansibler.utils.help import display_help


def main() -> None:
    """ Entry point for the script """
    run_ansibler()


def run_ansibler() -> None:
    """ Ansibler """
    # Get CMD args
    args = get_user_arguments()

    # Check for clear-cache
    if "clear-cache" in args:
        clear_cache()
        print("Cache cleared")

    json_file = args.get("json-file", "./ansibler.json")

    # Run generate compatibility charts
    if "generate-compatibility-chart" in args:
        molecule_results_dir = args.get("molecule-results-dir")
        generate_compatibility_chart(
            molecule_results_dir, json_file=json_file)
    elif "populate-platforms" in args:
        populate_platforms(json_file=json_file)
    elif "role-dependencies" in args:
        repository_status_template = args.get(
            "repository-status-template", None)
        variables = args.get("variables", None)

        # Read repo status template
        if repository_status_template is not None:
            template_is_file = check_file_exists(repository_status_template)
            if template_is_file:
                with open(repository_status_template) as f:
                    repository_status_template = f.read()
                    repository_status_template.replace("\n", "")

        # Read variables
        if variables is not None:
            variables_is_file = check_file_exists(variables)
            if not variables_is_file:
                print("Invalid variables file.")
                sys.exit(1)
            else:
                variables = read_json_file(variables)

        asyncio.run(
            generate_role_dependency_chart(
                json_file=json_file,
                template=repository_status_template,
                variables=variables
            )
        )
    else:
        if "clear-cache" not in args:
            display_help()


if __name__ == "__main__":
    main()



from unittest import TestCase
from unittest.mock import patch
from ansibler.role_dependencies.dependencies import (
    get_default_roles,
    parse_default_roles,
    get_role_dependency_link,
    get_role_dependency_description,
    get_role_dependency_supported_oses,
    get_role_dependency_status
)
from ansibler.exceptions.ansibler import CommandNotFound, RolesParseError


class TestRoleDependencies(TestCase):
    VALID_DEFAULT_ROLES = "DEFAULT_ROLES_PATH(/opt/Playbooks/ansible.cfg) = [" \
        "'/opt/Playbooks/roles/applications', '/opt/Playbooks/roles/crypto', " \
        "'/opt/Playbooks/roles/helpers', '/opt/Playbooks/roles/languages', '/" \
        "opt/Playbooks/roles/misc', '/opt/Playbooks/roles/services', '/opt/Pl" \
        "aybooks/roles/system', '/opt/Playbooks/roles/tools', '/opt/Playbooks" \
        "/roles/virtualization', '/root/.ansible/roles', '/usr/share/ansible/" \
        "roles', '/etc/ansible/roles']"

    @patch("ansibler.role_dependencies.dependencies.get_subprocess_output")
    def test_get_roles(self, mock_get_subprocess_output):
        """
        Test get roles

        Args:
            mock_get_subprocess_output (Mock): get_subprocess_output mock
        """
        mock_get_subprocess_output.return_value = "DEFAULT_ROLES_PATH (ansible)"
        roles = get_default_roles()
        self.assertEqual(roles, "DEFAULT_ROLES_PATH (ansible)")

    @patch("ansibler.role_dependencies.dependencies.get_subprocess_output")
    def test_get_roles_ansible_not_installed(self, mock_get_subprocess_output):
        """
        Makes sure an exception is raised when Ansible is not installed.

        Args:
            mock_get_subprocess_output (Mock): get_subprocess_output mock
        """
        mock_get_subprocess_output.return_value = "command not found"
        with self.assertRaises(CommandNotFound):
            _ = get_default_roles()

    def test_parse_default_roles(self):
        """
        Test parse default roles
        """
        roles = parse_default_roles(self.VALID_DEFAULT_ROLES)
        self.assertEqual(len(roles), 12)

    def test_parse_invalid_default_roles(self):
        """
        Makes sure an exception is raised when roles is improperly formatted
        """
        with self.assertRaises(RolesParseError):
            _ = parse_default_roles("invalid roles")

    def test_get_role_dependency_link(self):
        """
        Test get role dependency link
        """
        l = get_role_dependency_link({"role_name": "role", "namespace": "user"})

        expected = \
            f"<a href=\"https://galaxy.ansible.com/user/role\"" \
            f"title=\"user.role on Ansible Galaxy\" target=\"_" \
            f"blank\">user.role</a>"

        self.assertEqual(l, expected)

    def test_get_role_dependency_description(self):
        """
        Test get role dependency description
        """
        d = get_role_dependency_description({"description": "foo"})
        self.assertEqual(d, "foo")

    def test_get_role_dependency_supported_oses(self):
        """
        Test get role dependency supported OSs
        """
        metadata = {
            "platforms": [{"name": "Ubuntu"}, {"name": "MacOSX"}],
            "repository": "repo"
        }
        r = get_role_dependency_supported_oses(metadata)
        self.assertIn(
            "gitlab.com/megabyte-labs/assets/-/raw/master/icon/ubuntu.png", r)

    def test_get_role_dependency_supported_oses_invalid_os(self):
        """
        Makes sure get role dependency supported oses raises an exception when
        an invalid platform is read
        """
        metadata = {
            "platforms": [{"name": "InvalidOS"}],
            "repository": "repo"
        }
        with self.assertRaises(ValueError):
            _ = get_role_dependency_supported_oses(metadata)

    def test_get_role_dependency_status(self):
        """
        Test get role dependency status
        """
        metadata = {
            "role_name": "role",
            "namespace": "user",
            "repository": "foo",
            "repository_status": "bar"
        }
        r = get_role_dependency_status(metadata)

        expected = "<a href=\"foo\" title=\"user.role's repository\" target=" \
                   "\"_blank\"><img src=\"bar\" /></a>"

        self.assertEqual(r, expected)

    def test_get_role_dependency_status_unavaiable_repo(self):
        """
        Test get role dependency status when repo is unavailable (null)
        """
        metadata = {"repository_status": "foo"}
        r = get_role_dependency_status(metadata)
        self.assertEqual(r, "<img src=\"foo\" />")

    def test_get_role_dependency_status_unavailable_status(self):
        """
        Test get role dependency status when repo status is unavailable (null)
        """
        r = get_role_dependency_status({"repository_status": None})
        self.assertEqual(r, "Unavailable")
